package imageMedd.transformers.warpers;

import imageMedd.Pixel;
import imageMedd.transformers.TransformerVisitor;

import java.awt.image.BufferedImage;


public class FloydSteinbergDither extends Warper {
	private static final String key = "-floyd";
	protected int rbits, gbits, bbits;
	protected double twoUpR, twoUpB, twoUpG;
	private static double floyd3[][] = { { 0.0, 0.0, 0.0 },
			{ 0.0, 0.0, 0.375 }, { 0.0, 0.375, 0.25 } };

	public FloydSteinbergDither(int nbits) {
		super();
		setBitsNum(nbits);
	}

	public void setBitsNum(int nbits) {
		if (nbits == 1) {
			rbits = gbits = bbits = 0;
			return;
		} else if (nbits == 2) {
			rbits = 1;
			gbits = bbits = 0;
			return;
		}
		rbits = (nbits / 3 + (nbits % 3) - (nbits % 3 / 2));
		gbits = nbits / 3 + (nbits % 3) / 2;
		bbits = nbits / 3;
		twoUpR = Math.pow(2.0, rbits);
		twoUpB = Math.pow(2.0, rbits);
		twoUpG = Math.pow(2.0, bbits);
	}

	@Override
	public void setParam(int[] param) {
		if (param.length < 1)
			this.setBitsNum(1);
		else {
			if (param[0] < 1 || param[0] > 24)
				throw new RuntimeException("Invalid Floyd for this number of bits.");
			this.setBitsNum(param[0]);
		}
	}

	@Override
	public BufferedImage warp(BufferedImage image1) {
		BufferedImage result = new BufferedImage(image1.getWidth(), image1
				.getHeight(), BufferedImage.TYPE_INT_ARGB);
		// Copy source image to dest image
		for (int i = 0; i < image1.getWidth(); i++)
			for (int j = 0; j < image1.getHeight(); j++)
				result.setRGB(i, j, image1.getRGB(i, j));
		// Here begins Floyd - Steinberg Dithering Algorithm:
		for (int i = 0; i < result.getWidth(); i++)
			for (int j = 0; j < result.getHeight(); j++) {
				// For each pixel:
				Pixel c = new Pixel(result.getRGB(i, j));
				// Find closest color in colors available with existing bits
				Pixel newC = quantize(c);
				// Calculate each channel's error
				int rerr = c.getR() - newC.getR();
				int gerr = c.getG() - newC.getG();
				int berr = c.getB() - newC.getB();
				// set current pixel color as quantized color
				result.setRGB(i, j, newC.getRGB());
				// for each neighbor pixel:
				for (int k = -1; k < 2; k++)
					for (int l = -1; l < 2; l++) {
						if ((j == 0 && l == -1) || (i == 0 && k == -1)
								|| (i == result.getWidth() - 1 && k == 1)
								|| (j == result.getHeight() - 1 && l == 1))
							// If neighbor isn't in matrix
							continue;
						// What does algorithm add to this neighbor
						int rplus = (int) (rerr * floyd3[k + 1][l + 1]), gplus = (int) (gerr * floyd3[k + 1][l + 1]), bplus = (int) (berr * floyd3[k + 1][l + 1]);
						// distribute error to neighbor
						Pixel cur = new Pixel(result.getRGB(i + k, j + l));
						cur = new Pixel(
								Pixel.addComponent(cur.getR(), rplus), Pixel
										.addComponent(cur.getG(), gplus),
								Pixel.addComponent(cur.getB(), bplus),255);
						result.setRGB(i + k, j + l, cur.getRGB());
					}
			}
		return result;
	}

	protected static Pixel result = new Pixel(0x0); 
	protected Pixel quantize(Pixel c) {
		if (rbits + bbits + gbits == 0) {
			if (c.getR() + c.getB() + c.getG() > 127 * 3)
				result.setRGB(0xFFFFFFFF);
			else {
				result.setRGB(0xFF000000);
			}
			return result;
		} else if (rbits + bbits + gbits == 1) {
			int sum = c.getR() + c.getB() + c.getG();
			if (sum > 127 * 3) {
				if (sum > 191*3)
				result.setRGB(0xFFFFFFFF);
				else result.setRGB(0xFFAAAAAA);
			} else {
				if (sum < 64 * 3)
					result.setRGB(0xFF000000);
				else
					result.setRGB(0xFF555555);
			}
			return result;
		}
		int k = c.getR();
		k = k >> (8 - rbits);
		// k has the <bits> MSBs, it's a number between 0 to 2^bits-1
		int r = 255 * k / ((int) twoUpR - 1);
		k = c.getG();
		k = k >> (8 - gbits);
		int g = 255 * k / ((int) twoUpG - 1);
		k = c.getB();
		k = k >> (8 - bbits);
		int b = 255 * k / ((int) twoUpB - 1);
		result.setR(r);
		result.setG(g);
		result.setB(b);
		return result;
	}

	@Override
	public void accept(TransformerVisitor v) {
		v.visit(this);
	}

	@Override
	public String getString() {
		return key;
	}

	@Override
	public String help() {
		return "Performs Floyd Steinberg dither on the image. Takes number of color bits on parameter 0.";
	}
}
